<?php

namespace Drupal\monarch_ql\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\Sql\DefaultTableMapping;
use Drupal\Core\Entity\Sql\SqlContentEntityStorage;
use Drupal\Core\Entity\Sql\TableMappingInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\monarch_ql\Form\SettingsForm;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 * Returns responses for QL routes.
 */
final class QlController extends ControllerBase {

  /**
   * The field default filter list.
   */
  private array $queryFieldFilters = [];

  /**
   * Local cache for repeated actions.
   */
  private array $cache = [];

  /**
   * Settings.
   */
  private array $settings = [];

  /**
   * The entity_field.manager service.
   */
  private readonly EntityFieldManagerInterface $entityFieldManager;

  /**
   * The database service.
   */
  private readonly Connection $database;

  /**
   * The current user.
   */
  private readonly AccountInterface $account;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    $instance = new static();

    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->database = $container->get('database');
    $instance->moduleHandler = $container->get('module_handler');
    $instance->account = $container->get('current_user');
    $configFactory = $container->get('config.factory');

    if (file_exists(SettingsForm::SETTINGS_CACHE_FILE)) {
      $settings = json_decode(file_get_contents(SettingsForm::SETTINGS_CACHE_FILE), TRUE);
    }
    else {
      $settings = $configFactory->get('monarch_ql.settings')->get();
      file_put_contents(SettingsForm::SETTINGS_CACHE_FILE, json_encode($settings));
    }

    $instance->settings = $settings;

    return $instance;
  }

  /**
   * Check access.
   *
   * @todo Implement hooks.
   */
  public function access(Request $request = NULL, $query = NULL, $page = NULL, $count = NULL) : AccessResultInterface {
    $access = NULL;

    foreach (($this->moduleHandler->invokeAll('monarch_ql_access', [
      $this->account,
      $query,
      $page,
      $count,
      $request,
      $this->settings,
    ]) ?: []) as $v) {
      $access = ($access ?? $v) && $v;
    }

    return AccessResult::allowedIf($access ?? FALSE);
  }

  /**
   * Check entity type and field availability.
   */
  private function checkAvailable(array $field_access_list, string $entity_type, string $field_name = NULL) : bool {
    if ($field_name) {
      return (bool) ($field_access_list["$entity_type.$field_name"] ?? $field_access_list["$entity_type"] ?? NULL);
    }

    return (bool) ($field_access_list["$entity_type"] ?? NULL);
  }

  /**
   * Checks whether the given field can be stored in a shared table.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $storage_definition
   *   The field storage definition.
   *
   * @return bool
   *   TRUE if the field can be stored in a shared table, FALSE otherwise.
   */
  public function allowsSharedTableStorage(FieldStorageDefinitionInterface $storage_definition) {
    return !$storage_definition->hasCustomStorage() && $storage_definition->isBaseField() && !$storage_definition->isMultiple() && !$storage_definition->isDeleted();
  }

  /**
   * Checks whether the given field has to be stored in a dedicated table.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $storage_definition
   *   The field storage definition.
   *
   * @return bool
   *   TRUE if the field has to be stored in a dedicated table, FALSE otherwise.
   */
  public function requiresDedicatedTableStorage(FieldStorageDefinitionInterface $storage_definition) {
    return !$storage_definition->hasCustomStorage() && !$this->allowsSharedTableStorage($storage_definition);
  }

  /**
   * Get columns from a field storage definition.
   */
  public function getColumnNames(FieldStorageDefinitionInterface $storage_definition) : array {
    $_entity_type_id = $storage_definition->getTargetEntityTypeId();
    $_field_name = $storage_definition->getName();

    if (isset($this->cache[__FUNCTION__][$_entity_type_id][$_field_name])) {
      return $this->cache[__FUNCTION__][$_entity_type_id][$_field_name];
    }

    $ret = [];

    $field_name = $storage_definition->getName();

    if ($this->allowsSharedTableStorage($storage_definition)) {
      foreach (array_keys($storage_definition->getColumns()) as $property_name) {
        $ret[$property_name] = count($storage_definition->getColumns()) == 1 ? $field_name : $field_name . '__' . $property_name;
      }
    }
    elseif ($this->requiresDedicatedTableStorage($storage_definition)) {
      foreach (array_keys($storage_definition->getColumns()) as $property_name) {
        if ($property_name === TableMappingInterface::DELTA) {
          $ret[$property_name] = 'delta';
        }
        elseif ($property_name === 'deleted') {
          $ret[$property_name] = $property_name;
        }
        else {
          $ret[$property_name] = $field_name . '_' . $property_name;
        }
      }
    }

    return $this->cache[__FUNCTION__][$_entity_type_id][$_field_name] = $ret;
  }

  /**
   * Get the field table map.
   */
  public function getFieldTableMap(ContentEntityTypeInterface $entity_type) : array {
    $_entity_type_id = $entity_type->id();

    if (isset($this->cache[__FUNCTION__][$_entity_type_id])) {
      return $this->cache[__FUNCTION__][$_entity_type_id];
    }

    $ret = [];
    $storage = $this->entityTypeManager->getStorage($entity_type->id());

    if ($storage instanceof SqlContentEntityStorage) {
      $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type->id());
      $table_mapping = $storage->getTableMapping();

      if ($table_mapping instanceof DefaultTableMapping) {
        foreach (array_keys($field_storage_definitions) as $field_name) {
          $table_name = $table_mapping->getFieldTableName($field_name);
          $ret[$field_name] = $table_name;
        }
      }

      $r1 = $entity_type->getRevisionTable();
      $r2 = $entity_type->getRevisionDataTable();

      $ret = array_filter($ret, function ($table_name) use ($r1, $r2) {
        return $table_name !== $r1 && $table_name !== $r2;
      });
    }

    return $this->cache[__FUNCTION__][$_entity_type_id] = $ret;
  }

  /**
   * Convert hex keys back to the original keys.
   */
  private function convertKeysFromHex(array $q) : array {
    $ret = [];

    foreach ($q as $key => $value) {
      $ret[urldecode(hex2bin($key))] = is_array($value) ? $this->convertKeysFromHex($value) : $value;
    }

    return $ret;
  }

  /**
   * Parses a query string trying to preserve period and space in key names.
   */
  private function parseQueryString(string $str) : array {
    $parts = explode('&', $str);

    foreach ($parts as $index => $part) {
      $subparts = explode('=', $part);

      if (!is_null($subparts[0])) {
        $p = preg_split('/([\[\]])/', $subparts[0], -1, PREG_SPLIT_DELIM_CAPTURE);

        foreach ($p as $i => $s) {
          if ($s !== '[' && $s !== ']') {
            $p[$i] = bin2hex($s);
          }
        }

        $subparts[0] = implode('', $p);
      }

      $parts[$index] = implode('=', $subparts);
    }

    parse_str(implode('&', $parts), $q);

    return $this->convertKeysFromHex($q);
  }

  /**
   * Flatten array.
   */
  private function flattenWithoutKeys(array $arr, array &$ret = []) : array {
    foreach ($arr as $value) {
      if (is_array($value)) {
        $this->flattenWithoutKeys($value, $ret);
      }
      else {
        $ret[] = $value;
      }
    }

    return $ret;
  }

  /**
   * Builds the response.
   */
  public function build(Request $request = NULL, $query = NULL, $page = NULL, $count = NULL) : Response {
    $start = hrtime(TRUE);

    $request_variables = [
      'request' => $request,
      'query' => $query,
      'page' => $page,
      'count' => $count,
      'account' => $this->account,
      'settings' => $this->settings,
    ];

    $json = [];

    if (is_null($query)) {
      $json['available'] = [];

      foreach (($this->settings['fields'] ?? []) as $field => $value) {
        if ($value) {
          [$entity_type_id, $field_name] = explode('|', $field, 2) + [NULL, NULL];
          $json['available'][$entity_type_id][] = $field_name;
        }
      }

      foreach ($json['available'] as $entity_type_id => &$fields) {
        sort($fields);
      }

      ksort($json['available']);

      return new JsonResponse((object) $json, 200, [
        'x-elapsed' => (hrtime(TRUE) - $start) / 1000000.0,
      ]);
    }

    $field_access_list = [];

    $system_filter_list = $this->queryFieldFilters;
    $user_filter_list = $this->parseQueryString(explode('?', $request->getRequestUri(), 2)[1] ?? '');

    foreach ($user_filter_list as $key => $value) {
      if (is_array($value)) {
        $user_filter_list[$key] = $this->flattenWithoutKeys($value);
      }
    }

    $definitions = array_filter($this->entityTypeManager->getDefinitions(), function ($type) {
      return $type instanceof ContentEntityTypeInterface;
    });

    foreach ($definitions as $entity_type_id => $definition) {
      if ($pub_key = $definition->getKey('published') ?: $definition->getKey('status')) {
        $system_filter_list["$entity_type_id.$pub_key !="] = 0;
      }
      else {
        $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);

        if ($field_storage_definitions['status'] ?? NULL) {
          $system_filter_list["$entity_type_id.status !="] = 0;
        }
      }
    }

    $settings = $this->settings;

    $this->moduleHandler->alter('monarch_ql_query', $field_access_list, $system_filter_list, $settings);
    $this->moduleHandler->alter('monarch_ql_field_access', $field_access_list, $settings);
    $this->moduleHandler->alter('monarch_ql_field_filters', $system_filter_list, $settings);

    $page = max(0, intval($page ?? 0));
    $count = max(1, min(SettingsForm::ABSOLUTE_MAX, intval($settings['max'] ?? 500), intval($count ?? $settings['count'] ?? 25)));

    $query_list = [];

    foreach (explode(' ', $query) as $query_part) {
      [$entity_type_id, $field_name, $column] = explode('.', $query_part, 3) + [NULL, NULL, NULL];

      $definition = $definitions[$entity_type_id] ?? NULL;

      if (!$definition || !$definition->getBaseTable() || !($definition instanceof ContentEntityTypeInterface)) {
        continue;
      }

      $field_table_map = $this->getFieldTableMap($definition);

      $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);
      $id_key = $definition->getKey('id');

      if ($field_name) {
        if ($this->checkAvailable($field_access_list, $entity_type_id, $field_name)) {
          $query_list[$entity_type_id][$definition->getBaseTable()] = $query_list[$entity_type_id][$definition->getBaseTable()] ?? [];
        }
        else {
          continue;
        }

        $query_list[$entity_type_id][$definition->getBaseTable()] = $query_list[$entity_type_id][$definition->getBaseTable()] ?? [];

        if ($field_storage_definition = $field_storage_definitions[$field_name] ?? NULL) {
          $columns = $this->getColumnNames($field_storage_definition);
          $table_name = $field_table_map[$field_name] ?? NULL;

          if (!$table_name) {
            continue;
          }

          if ($column) {
            if ($columns[$column] ?? NULL) {
              $query_list[$entity_type_id][$table_name][$field_name][$column] = $columns[$column];
            }
          }
          else {
            foreach (array_keys($columns) as $column) {
              $query_list[$entity_type_id][$table_name][$field_name][$column] = $columns[$column];
            }
          }
        }
      }
      else {
        foreach ($field_storage_definitions as $field_name => $field_storage_definition) {
          if ($this->checkAvailable($field_access_list, $entity_type_id, $field_name)) {
            $query_list[$entity_type_id][$definition->getBaseTable()] = $query_list[$entity_type_id][$definition->getBaseTable()] ?? [];
          }
          else {
            continue;
          }

          $table_name = $field_table_map[$field_name] ?? NULL;

          if (!$table_name) {
            continue;
          }

          $columns = $this->getColumnNames($field_storage_definition);

          foreach (array_keys($columns) as $column) {
            $query_list[$entity_type_id][$table_name][$field_name][$column] = $columns[$column];
          }
        }
      }
    }

    $filter_list = [];

    foreach ([
      $system_filter_list,
      $user_filter_list,
    ] as $sublist) {
      foreach ($sublist as $filter => $value) {
        [$filter, $operator] = explode(' ', $filter, 2) + [NULL, NULL];
        [$entity_type_id, $field_name, $column] = explode('.', $filter, 3) + [NULL, NULL, NULL];

        if (!isset($query_list[$entity_type_id]) || empty($field_name)) {
          continue;
        }

        if (!$this->checkAvailable($field_access_list, $entity_type_id, $field_name)) {
          continue;
        }

        $field = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id)[$field_name] ?? NULL;

        if (!$field) {
          continue;
        }

        if (empty($column)) {
          $column = $field->getMainPropertyName();
        }

        $columns = $this->getColumnNames($field);
        $column = $columns[$column] ?? NULL;

        if (!$column) {
          continue;
        }

        $table_name = $field_table_map[$field_name] ?? NULL;

        if (!$table_name) {
          continue;
        }

        if ($operator) {
          $filter_list[$entity_type_id][$table_name][$column][] = [$value, $operator];
        }
        else {
          if (is_array($value)) {
            $filter_list[$entity_type_id][$table_name][$column][] = [$value, 'IN'];
          }
          else {
            $filter_list[$entity_type_id][$table_name][$column][] = [$value, '='];
          }
        }
      }
    }

    foreach ($query_list as $entity_type_id => $query_tables) {
      $definition = $definitions[$entity_type_id];
      $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);
      $id_key = $definition->getKey('id');

      $ids = NULL;

      foreach ($query_tables as $query_table => $query_fields) {
        $query = $this->database->select($query_table, 't');
        $real_id_key = $query_table === $definition->getBaseTable() || $query_table === $definition->getDataTable() ? $id_key : 'entity_id';
        $query->addField('t', $real_id_key);

        foreach ($query_fields as $query_field => $columns) {
          foreach ($columns as $column) {
            if ($column !== $real_id_key) {
              $query->addField('t', $column);
            }
          }
        }

        if ($query_table === $definition->getBaseTable()) {
          if ($query_table === 'file_managed') {
            $schemes = $query->andConditionGroup()
              ->condition('t.uri', 'private://%', 'NOT LIKE')
              ->condition('t.uri', 'temporary://%', 'NOT LIKE');

            $file_conditions = $query->orConditionGroup();
            $file_conditions->condition('t.uid', $this->account->isAuthenticated() ? $this->account->id() : -1);

            $file_conditions->condition($schemes);
            $query->condition($file_conditions);
          }

          foreach ($filter_list[$entity_type_id] ?? [] as $filter_table => $filter_columns) {
            $filter_query = $this->database->select($filter_table, 'f');
            $filter_id_key = $filter_table === $definition->getBaseTable() || $filter_table === $definition->getDataTable() ? $id_key : 'entity_id';
            $filter_query->addField('f', $filter_id_key, 'id');

            foreach ($filter_columns as $filter_column => $filter_params_list) {
              foreach ($filter_params_list as $filter_params) {
                $filter_query->condition($filter_column, ...$filter_params);
              }
            }

            $query->condition($real_id_key, $filter_query, 'IN');
          }

          foreach ($query->orderBy($id_key)->range($page * $count, $count)->execute() as $row) {
            foreach ($query_fields as $query_field => $columns) {
              $field_storage_definition = $field_storage_definitions[$query_field] ?? NULL;

              foreach ($columns as $column => $row_key) {
                $column_type = $field_storage_definition->getColumns()[$column]['type'];

                $value = ($row->{$row_key} ?? NULL);

                if (($column_type === 'int' || $column_type === 'float' || $column_type === 'numeric') && is_numeric($value)) {
                  $value = floatval($value);
                }

                if (count($field_storage_definition->getColumns()) === 1) {
                  $json[$entity_type_id][$row->{$real_id_key}][$query_field] = $value;
                }
                else {
                  $json[$entity_type_id][$row->{$real_id_key}][$query_field][$column] = $value;
                }
              }
            }

            $ids[] = $row->{$real_id_key};
          }
        }
        elseif ($ids) {
          if ($query_table === $definition->getDataTable()) {
            foreach ($query->condition($real_id_key, $ids, 'IN')->execute() as $row) {
              foreach ($query_fields as $query_field => $columns) {
                $field_storage_definition = $field_storage_definitions[$query_field] ?? NULL;

                foreach ($columns as $column => $row_key) {
                  $column_type = $field_storage_definition->getColumns()[$column]['type'];

                  $value = ($row->{$row_key} ?? NULL);

                  if (($column_type === 'int' || $column_type === 'float' || $column_type === 'numeric') && is_numeric($value)) {
                    $value = floatval($value);
                  }

                  if (count($field_storage_definition->getColumns()) === 1) {
                    $json[$entity_type_id][$row->{$real_id_key}][$query_field] = $value;
                  }
                  else {
                    $json[$entity_type_id][$row->{$real_id_key}][$query_field][$column] = $value;
                  }
                }
              }
            }
          }
          else {
            $query->addField('t', 'delta');

            foreach ($query->condition($real_id_key, $ids, 'IN')->execute() as $row) {
              foreach ($query_fields as $query_field => $columns) {
                $field_storage_definition = $field_storage_definitions[$query_field] ?? NULL;

                foreach ($columns as $column => $row_key) {
                  $column_type = $field_storage_definition->getColumns()[$column]['type'];

                  $value = ($row->{$row_key} ?? NULL);

                  if (($column_type === 'int' || $column_type === 'float' || $column_type === 'numeric') && is_numeric($value)) {
                    $value = floatval($value);
                  }

                  if ($field_storage_definition->getCardinality() === 1) {
                    if (count($field_storage_definition->getColumns()) === 1) {
                      $json[$entity_type_id][$row->{$real_id_key}][$query_field] = $value;
                    }
                    else {
                      $json[$entity_type_id][$row->{$real_id_key}][$query_field][$column] = $value;
                    }
                  }
                  else {
                    if (count($field_storage_definition->getColumns()) === 1) {
                      $json[$entity_type_id][$row->{$real_id_key}][$query_field][$row->delta] = $value;
                    }
                    else {
                      $json[$entity_type_id][$row->{$real_id_key}][$query_field][$row->delta][$column] = $value;
                    }
                  }
                }
              }
            }
          }
        }
      }

      if (is_array($json[$entity_type_id] ?? NULL)) {
        $json[$entity_type_id] = array_values($json[$entity_type_id]);
      }
    }

    $headers = [
      'x-ql-elapsed' => (hrtime(TRUE) - $start) / 1000000.0,
    ];

    $this->moduleHandler->alter('monarch_ql_json', $json, $headers, $request_variables);

    return new JsonResponse((object) $json, 200, $headers);
  }

}
