<?php

namespace Drupal\monarch_ql\Form;

use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Sql\SqlContentEntityStorage;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure Monarch QL settings for this site.
 */
final class SettingsForm extends ConfigFormBase {

  const SETTINGS_CACHE_FILE = 'public://ql_settings.json';
  const ABSOLUTE_MAX = 10000;

  /**
   * The entity_type.manager service.
   */
  private EntityTypeManagerInterface $entityTypeManager;

  /**
   * The entity_field.manager service.
   */
  private EntityFieldManagerInterface $entityFieldManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);

    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->entityFieldManager = $container->get('entity_field.manager');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'monarch_ql_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return ['monarch_ql.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $form['count'] = [
      '#type' => 'number',
      '#title' => $this->t('Default Results per Page'),
      '#min' => 1,
      '#max' => static::ABSOLUTE_MAX,
      '#default_value' => max(1, min(static::ABSOLUTE_MAX, $this->config('monarch_ql.settings')->get('count') ?? 25)),
    ];

    $form['max'] = [
      '#type' => 'number',
      '#title' => $this->t('Max Allowed Results per Page'),
      '#min' => 1,
      '#max' => static::ABSOLUTE_MAX,
      '#default_value' => max(1, min(static::ABSOLUTE_MAX, $this->config('monarch_ql.settings')->get('max') ?? 500)),
    ];

    $form['owned'] = [
      '#type' => 'checkboxes',
      '#options' => [],
      '#title' => $this->t('Query Only Owned Entities'),
      '#default_value' => $this->config('monarch_ql.settings')->get('owned') ?? [],
    ];

    $form['fields'] = [
      '#type' => 'checkboxes',
      '#options' => [],
      '#title' => $this->t('Fields Allowed in Query or Filter'),
      '#default_value' => $this->config('monarch_ql.settings')->get('fields') ?? [],
    ];

    $definitions = ($this->entityTypeManager->getDefinitions() ?: []);

    ksort($definitions);

    foreach ($definitions as $entity_type_id => $definition) {
      if (!$definition || !$definition->getBaseTable() || !($definition instanceof ContentEntityTypeInterface)) {
        continue;
      }

      $storage = $this->entityTypeManager->getStorage($entity_type_id);

      if ($storage instanceof SqlContentEntityStorage) {
        $field_storage_definitions = ($this->entityFieldManager->getFieldStorageDefinitions($entity_type_id) ?: []);

        $owner_key = $definition->getKey('owner') ?: $definition->getKey('user') ?: $definition->getKey('uid');

        if ($owner_key || $entity_type_id === 'user') {
          $form['owned']['#options'][$entity_type_id] = $this->t('@entity', [
            '@entity' => $entity_type_id,
          ]);
        }

        $table_mapping = $storage->getTableMapping();

        foreach (array_keys($field_storage_definitions) as $field_name) {
          $table_name = $table_mapping->getFieldTableName($field_name);
          if ($table_name !== $definition->getRevisionTable() && $table_name !== $definition->getRevisionDataTable()) {
            $form['fields']['#options'][$entity_type_id . '|' . $field_name] = $this->t('@entity.<strong>@name</strong>', [
              '@entity' => $entity_type_id,
              '@name' => $field_name,
            ]);
          }
        }
      }
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $config = $this->config('monarch_ql.settings');

    $config
      ->set('count', $form_state->getValue('count'))
      ->set('max', $form_state->getValue('max'))
      ->set('owned', $form_state->getValue('owned'))
      ->set('fields', $form_state->getValue('fields'))
      ->save();

    $settings = $config->get();
    file_put_contents(SettingsForm::SETTINGS_CACHE_FILE, json_encode($settings));

    parent::submitForm($form, $form_state);
  }

}
